#!/usr/bin/env python3

# (c) Copyright 2022 CORSIKA Project, corsika-project@lists.kit.edu
#
# See file AUTHORS for a list of contributors.
#
# This software is distributed under the terms of the GNU General Public
# Licence version 3 (GPL Version 3). See file LICENSE for a full version of
# the license.


import pickle, sys, itertools



def load_particledb(filename):
    '''
    loads the pickled particle_db (which is an OrderedDict)
    '''
    with open(filename, "rb") as f:
        particle_db = pickle.load(f)
    return particle_db



def read_sophia_codes(filename, particle_db):
    '''
    reads to sophia codes data file

    For particls known to sophia, add 'sophia_code' and 'sophia_xsType' to particle_db
    '''
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if len(line)==0 or line[0] == '#':
                continue            
            identifier, sib_code, canInteractFlag, xsType = line.split()
            try:
                particle_db[identifier]["sophia_code"] = int(sib_code)
                particle_db[identifier]["sophia_xsType"] = xsType
            except KeyError as e:
                raise Exception("Identifier '{:s}' not found in particle_db".format(identifier))


            

def generate_sophia_enum(particle_db):
    '''
     generates the enum to access sophia particles by readable names
    '''
    output = "enum class SophiaCode : int8_t {\n"
    for identifier, pData in particle_db.items():
        if 'sophia_code' in pData:
            output += "  {:s} = {:d},\n".format(identifier, pData['sophia_code'])
    output += "};\n"
    return output



def generate_corsika2sophia(particle_db):    
    '''
    generates the look-up table to convert corsika codes to sophia codes
    '''
    string = "std::array<SophiaCode, {:d}> constexpr corsika2sophia = {{\n".format(len(particle_db))
    for identifier, pData in particle_db.items():
        if pData['isNucleus']: continue
        if 'sophia_code' in pData:
            string += "  SophiaCode::{:s}, \n".format(identifier)
        else:
            string += "  SophiaCode::Unknown, // {:s}\n".format(identifier + ' not implemented in sophia')
    string += "};\n"
    return string
    


def generate_corsika2sophia_xsType(particle_db):    
    '''
    generates the look-up table to convert corsika codes to sophia codes
    '''
    string = "std::array<SophiaXSClass, {:d}> constexpr corsika2sophiaXStype = {{\n".format(len(particle_db))
    for identifier, pData in particle_db.items():
        if pData['isNucleus']: continue
        if 'sophia_xsType' in pData:
            string += "  SophiaXSClass::{:s}, // {:s}\n".format(pData['sophia_xsType'], identifier)
        else:
            string += "  SophiaXSClass::CannotInteract, // {:s}\n".format(identifier + ' not implemented in sophia')
    string += "};\n"
    return string


def generate_sophia2corsika(particle_db) :
    '''
    generates the look-up table to convert sophia codes to corsika codes    
    '''
    string = ""
    
    minID = 0
    for identifier, pData in particle_db.items() :
        if 'sophia_code' in pData:
            minID = min(minID, pData['sophia_code'])

    string += "SophiaCodeIntType constexpr minSophia = {:d};\n\n".format(minID)

    pDict = {}
    for identifier, pData in particle_db.items() :
        if 'sophia_code' in pData:
            sib_code = pData['sophia_code'] - minID
            pDict[sib_code] = identifier
    
    nPart = max(pDict.keys()) - min(pDict.keys()) + 1
    string += "std::array<corsika::Code, {:d}> constexpr sophia2corsika = {{\n".format(nPart)
    
    for iPart in range(nPart) :
        if iPart in pDict:
            identifier = pDict[iPart]
        else:
            identifier = "Unknown"
        string += "  corsika::Code::{:s}, \n".format(identifier)
    
    string += "};\n"
    return string

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("usage: {:s} <particle_db.pkl> <sophia_codes.dat>".format(sys.argv[0]), file=sys.stderr)
        sys.exit(1)
        
    print("code_generator.py for sophia")
    
    particle_db = load_particledb(sys.argv[1])
    read_sophia_codes(sys.argv[2], particle_db)
    
    with open("Generated.inc", "w") as f:
        print("// this file is automatically generated\n// edit at your own risk!\n", file=f)
        print(generate_sophia_enum(particle_db), file=f)
        print(generate_corsika2sophia(particle_db), file=f)
        print(generate_sophia2corsika(particle_db), file=f)
        #print(generate_corsika2sophia_xsType(particle_db), file=f)
